/*
   FILE:							       	
	idle_analyzer.c
								       	
   DESCRIPTION:						
	      	
	See README file for details.
					       				
   Copyright (C) 2001
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        March 2004

*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "./common.h"


#include <trace.h>

#define _DEBUG

#define MAX_DATA_SIZE 32
#define MAX_HYPERPERIODS 1000


struct rates
{
  int n;
  long long th;
  long long max;
  long long min;
  long long tot;
  double sdiff;
  double avg;
  double var;
};

static struct rates idle_time_rate;



static void inline
initialize_figures (struct rates *figures) {
  figures->max   = (long) 0;
  figures->min   = (long) 1000000000;
  figures->tot   = (long) 0;
  figures->sdiff = (long) 0;
}


static void inline
store_figures (struct rates *figures, long long value)
{

  figures->n++;
  figures->tot += value;
  if (figures->max < value)
    {
      figures->max = value;
    }
  if (figures->min > value)
    {
      figures->min = value;
    }
  figures->avg = (double) ((double) figures->tot / (double) figures->n);
}





static void inline
calculate_square_diffs (struct rates *figures, long long value)
{
  double square_diff;

  square_diff = ((double) value) - figures->avg;
  square_diff *= square_diff;
  figures->sdiff += square_diff;
}




static void inline
calculate_variance (struct rates *figures)
{
  if (figures->n > 0)
    {
      figures->var = ((double) figures->sdiff) / ((double) figures->n);
    }
}




void
present_results (struct rates *figures)
{

  fprintf (stdout, "\n\n");
  fprintf (stdout,
	   " ,--------------------------------------------------------------------------------------------.\n");
  fprintf (stdout,
	   " |           STATISTICAL    RESULTS     (all values in nsec)                                  |\n");
  fprintf (stdout,
	   " +-------------------+------------------------------------------------------------------------+\n");
  fprintf (stdout,
	   " |  MEASURED VALUESS |    N        MIN           MAX            AVG               VAR         |\n");
  fprintf (stdout,
	   " +-------------------+------------------------------------------------------------------------+\n");
  fprintf (stdout, " |  %s  |%6d  %12lld  %12lld     %12.4f    %12.4f  |\n",
	   "Idle time      ",
	   figures->n,
	   figures->min,
	   figures->max,
	   figures->avg, 
	   figures->var);
  fprintf(stdout, 
	  " `--------------------------------------------------------------------------------------------'\n");
  
  fprintf (stdout, "\n\n");

}




int
main (int argc, char *argv[])
{

  char log_filename[256];
  char data[MAX_DATA_SIZE];
  char str[TRACE_NAME_MAX];
  size_t data_len;
  int unavailable;
  int error;
  int log_fd;

  trace_id_t trace_id;
  struct posix_trace_event_info event;
  struct idle_measurement_data *idle_data;

  struct timespec first_release;
  struct timespec hyperperiod;
  struct timespec next_hyperperiod;
  struct timespec start_linux_interval;
  struct timespec linux_interval;
  //  struct timespec difference;
  struct timespec idle_average;

  struct timespec idle_time[MAX_HYPERPERIODS];
  int             n_events[MAX_HYPERPERIODS];
  int n_hyperperiod = 0;
  int start_counting = 0;
  int info_read = 0;
  int i;
  long long nsec;
  //  long long avg_nsec;
  //  long      diffs;

  if (argc < 2)
    {
      fprintf (stderr, "Use: linux_analyzer logfile\n");
      exit (1);
    }
  else if (argc == 2)
    {
      strncpy (log_filename, argv[1], sizeof (log_filename));
    }


  if ((log_fd = open (log_filename, O_RDONLY)) < 0)
    {
      fprintf (stderr, "linux_analyzer: Error while opening log file %s\n",
	       log_filename);
      exit (1);
    }

  error = posix_trace_open (log_fd, &trace_id);
  if (error < 0)
    {
      fprintf (stderr,
	       "linux_analyzer: Error while opening pre-recorded stream\n");
      exit (1);

    }

  // Initialize data structures:
  for (i=0; i<MAX_HYPERPERIODS; i++) {
    idle_time[i].tv_sec = 0L;
    idle_time[i].tv_nsec = 0L;
    n_events[i] = 0;
  }

  // Start retrieving events while the stream is open: 
  while (1)
    {

      error =
	posix_trace_getnext_event (trace_id, &event, &data, sizeof (data),
				   &data_len, &unavailable);

      if (error)
	{
	  fprintf (stdout, "Error (%d)\n", error);
	  break;
	}

      if (unavailable)
	{
	  //fprintf (stdout, "Event unavailable\n");
	  break;
	}

      posix_trace_eventid_get_name (trace_id, event.posix_event_id, str);

#ifdef _DEBUG
      fprintf(stderr, "(%ld.%09ld) Event %d: %s\n", 
	      event.posix_timestamp.tv_sec, 
	      event.posix_timestamp.tv_nsec,
	      event.posix_event_id, str);
#endif

      if (start_counting) {
	n_events[n_hyperperiod]++;
      }

      if (info_read)
	{
	  if (!start_counting && timespec_ge (&event.posix_timestamp,
					      &first_release))
	    {
	      //idle_time[0].tv_sec = 0L;
	      //idle_time[0].tv_nsec = 0L;
	      start_linux_interval = first_release;
	      start_counting = 1;
#ifdef _DEBUG
      fprintf(stderr, 
	      ">>>>>>>> START COUNTING >>>>>>>>>>>>>>>>> Hyp (%ld.%09ld) \n",
	      first_release.tv_sec,
	      first_release.tv_nsec);
#endif
	    }
	}


      if (!strcmp (str, "idle-measurement-info"))
	{
	  idle_data = (struct idle_measurement_data *) data;
	  first_release = idle_data->first_release;
	  hyperperiod = idle_data->hyperperiod;
	  next_hyperperiod = first_release;
	  timespec_add (&next_hyperperiod, &hyperperiod);
	  info_read = 1;
#ifdef _DEBUG
	  fprintf (stdout, "[INFO] first = %ld.%09ld   hyp = %ld.%09ld\n",
		   first_release.tv_sec, first_release.tv_nsec,
		   hyperperiod.tv_sec, hyperperiod.tv_nsec);
	  fprintf (stdout, "[RUN] next_hyperperiod = %ld.%09ld\n",
		   next_hyperperiod.tv_sec, next_hyperperiod.tv_nsec);
#endif
	}

      else if (!strcmp (str, "posix_trace_cswitch_tolinux"))
	{
	  start_linux_interval = event.posix_timestamp;
	}

      else if (!strcmp (str, "posix_trace_cswitch_totask"))
	{
	  if (start_counting)
	    {
	      if (timespec_lt (&event.posix_timestamp, &next_hyperperiod))
		{
		  linux_interval = event.posix_timestamp;
		  timespec_sub (&linux_interval, &start_linux_interval);
		  timespec_add (&idle_time[n_hyperperiod], &linux_interval);
		}
	      else
		{
		  linux_interval = next_hyperperiod;
		  timespec_sub (&linux_interval, &start_linux_interval);
		  timespec_add (&idle_time[n_hyperperiod], &linux_interval);
		  n_hyperperiod++;
#ifdef _DEBUG
      fprintf(stderr, 
	      ">>>>>>>> NEXT HYPERPERIOD >>>>>>>>>>>>>>> Hyp (%ld.%09ld) <\n",
	      next_hyperperiod.tv_sec,
	      next_hyperperiod.tv_nsec);
#endif
		  linux_interval = event.posix_timestamp;
		  timespec_sub (&linux_interval, &next_hyperperiod);
		  timespec_add (&idle_time[n_hyperperiod], &linux_interval);
		  timespec_add (&next_hyperperiod, &hyperperiod);
#ifdef _DEBUG
		  fprintf (stdout, "[RUN] next_hyperperiod = %ld.%09ld\n",
			   next_hyperperiod.tv_sec, next_hyperperiod.tv_nsec);
#endif
		}
	    }
	}
    }


  // Statistical analysis of the idle time per hyperperiod:

  initialize_figures(&idle_time_rate);

  for (i = 0; i < n_hyperperiod; i++)
    {
#ifdef _DEBUG
      fprintf (stdout, "Idle time [%d]:  %ld.%09ld\n", i, idle_time[i].tv_sec,
	       idle_time[i].tv_nsec);
#endif

      nsec = (long long) (((long long) idle_time[i].tv_sec * 1000000000L) +
			  (long long) idle_time[i].tv_nsec);
      store_figures (&idle_time_rate, nsec);

    }

  // Now the average is calculated:
  idle_average.tv_sec  = (long) ((long long) idle_time_rate.avg / 1000000000);
  idle_average.tv_nsec = (long) ((long long) idle_time_rate.avg % 1000000000);
  //fprintf (stdout, "%ld.%09ld\n", idle_average.tv_sec, idle_average.tv_nsec);
		

  for (i = 0; i < n_hyperperiod; i++)
    {
      nsec = (long long) (((long long) idle_time[i].tv_sec * 1000000000L) +
			  (long long) idle_time[i].tv_nsec);
      calculate_square_diffs (&idle_time_rate, nsec);
    }

  calculate_variance (&idle_time_rate);

  present_results (&idle_time_rate);


  for (i = 0; i < n_hyperperiod; i++) {

    //      fprintf (stderr, "%ld.%09ld\n", idle_time[i].tv_sec,
    //       idle_time[i].tv_nsec);
      fprintf (stderr, "%i \t%d\n", i, n_events[i]);

  }

  return 0;
}
