/*

   FILE:
	posix_trace.c
       
   DESCRIPTION:							       	
        Linux side of the POSIX 1003.1q tracing implementation in RT-Linux 
        (v 3.0)
					       				
   Copyright (C) 2001
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        December 2003

*/


#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>

#include <trace.h>
#include <posix/trace/comm.h>


/* 
   GENERAL NOTE:

   This file is compiled into a library which includes all the POSIX
   trace support for Linux processes that we currently offer. In
   particular, there are two different kinds of support:

    1) The support for opening a log file into a pre-recorded stream
       and retrieving all the information (_POSIX_TRACE_LOGRET)

    2) The support for tracing events into an active stream which has
       to be created in RT-Linux (_POSIX_TRACE_LINUXTRACE). Obviously,
       this option is subordinate to having the trace support
       available in RT-Linux. In particular, the _POSIX_TRACE_UEVENTS
       option, which is in turn subordinate to the _POSIX_TRACE_CORE.
       All these dependencies are dealt with when configuring the
       RT-Linux kernel (e.g., make xconfig)

*/

#if defined _POSIX_TRACE_LOGRET || defined _POSIX_TRACE_LINUXTRACE


#define _TRACE_DEBUG

#ifdef _TRACE_DEBUG
#include <errno.h>
#endif









/* 
   LIBRARY DEFINITIONS AND TYPES 
*/

#define TRACE_STREAM_MAGIC      0xaafd
#define TRACE_ATTR_MAGIC        0xabef
#define NO_MAGIC                0x0




#ifdef _POSIX_TRACE_LOGRET


/* 
   LIBRARY GLOBAL VARIABLES 
*/

static stream_event_t           event_read;
static event_type_info_t        event_types[TRACE_EVENT_MAX];
static trace_attr_t             trid_attr;
static posix_trace_status_info  trid_status;
static int                      current_event;
static int                      pre_trid = -1;
static off_t                    first_event_pos;




/* 
   LIBRARY GLOBAL MACROS 
*/


#define TRACE_ATTR_IS_VALID(attr) ( \
                                     attr->magic == TRACE_ATTR_MAGIC \
                                  )

#define TRACE_STREAM_IS_VALID(trid) ( \
                                     trid == pre_trid \
                                    )



/* 
   ---------------------------------------------------------------------
   24.3. IMPLEMENTATION OF THE TRACE FUNCTIONS
   ---------------------------------------------------------------------
*/

//
// 24.3.2. Attribute initialization and destruction
//

int posix_trace_attr_init (trace_attr_t * attr) {

  return -EINVAL;
}





int posix_trace_attr_destroy (trace_attr_t *attr) {

  return -EINVAL;
}





//
// 24.3.3. Retrieve and set information about a trace stream
//
int posix_trace_attr_getgenversion (const trace_attr_t *attr, 
				    char *genversion) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  strncpy(genversion, attr->generation_version, TRACE_NAME_MAX);
  return 0;
}





int posix_trace_attr_getname (const trace_attr_t *attr,
			      char *tracename) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  strncpy(tracename, attr->trace_name, TRACE_NAME_MAX);
  return 0;
}





int posix_trace_attr_setname (trace_attr_t *attr, 
			      const char *tracename) {

    return -EINVAL;
}





int posix_trace_attr_getcreatetime (const trace_attr_t *attr, 
				    struct timespec *createtime) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *createtime = attr->creation_time;
  return 0;
}





int posix_trace_attr_getclockres (const trace_attr_t *attr,
				  struct timespec * resolution) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *resolution = attr->clock_resolution;
  return 0;
}





//
// 24.3.4. Retrieve and set the behavior of a trace stream
//
int posix_trace_attr_getstreamfullpolicy (const trace_attr_t *attr, 
					  int *streampolicy) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *streampolicy  = attr->stream_full_policy;
  return 0;
}





//
// 24.3.5. Retrieve and set trace stream size attributes
//
int posix_trace_attr_getmaxusereventsize (const trace_attr_t *attr, 
					   size_t data_len, 
					   size_t *eventsize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *eventsize= attr->header_size + attr->data_size - sizeof(void *);
  return 0;
}





int posix_trace_attr_getmaxsystemeventsize (const trace_attr_t *attr,
					    size_t *eventsize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *eventsize= attr->header_size + attr->data_size - sizeof(void *);
  return 0;
}





int posix_trace_attr_getmaxdatasize (const trace_attr_t *attr, 
				     size_t *maxdatasize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *maxdatasize = attr->max_data_size;
  return 0;
}





int posix_trace_attr_setmaxdatasize (trace_attr_t *attr, 
				     size_t maxdatasize) {

  /* Not supported in Linux */
  return -EINVAL;
}





int posix_trace_attr_getstreamsize (const trace_attr_t *attr, 
				    size_t *streamsize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }
 
  *streamsize = attr->stream_min_size;
  return 0;
}





int posix_trace_attr_setstreamsize (trace_attr_t *attr, 
				    size_t streamsize) {

  /* Not supported in Linux */
  return -EINVAL;
}





//
// 24.3.6. Trace stream initialization, flush and shutdown
//
int posix_trace_create (pid_t pid,
			const trace_attr_t *attr,
			trace_id_t *trid) {
    return -EINVAL;
}





int posix_trace_shutdown (trace_id_t trid) {

  return -EINVAL;
}






//
// 24.3.7. Clear trace stream and trace log
//
extern int posix_trace_clear (trace_id_t trid) {

  /* Not supported in Linux */
  return -EINVAL;
}






//
// 24.3.8. Manipulate trace event type identifier
//
#ifdef _POSIX_TRACE_EVENT_FILTER
int posix_trace_trid_eventid_open (trace_id_t trid,
				   const char *event_name,
				   trace_event_id_t *event) {

  /* Not supported in Linux */
  return -EINVAL; 
}
#endif 






extern int posix_trace_eventid_get_name (trace_id_t trid,
					 trace_event_id_t event,
					 char *event_name) {

  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
    return -EINVAL;
  }

  if (event_types[event].eventid < 0) {
    return -EINVAL;
  }

  strncpy(event_name, event_types[event].name, TRACE_EVENT_NAME_MAX);
  return 0;

}





int posix_trace_eventid_equal (trace_id_t trid,
			       trace_event_id_t event1,
			       trace_event_id_t event2) {

  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
    return -EINVAL;
  }

  
  return (event1 == event2);
}





//
// 24.3.9. Iterate over a mapping of trace event types
//
int posix_trace_eventtypelist_getnext_id (trace_id_t trid,
					  trace_event_id_t *event,
					  int *unavailable) {
  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
    return -EINVAL;
  }

  while (current_event < TRACE_EVENT_MAX && 
	 event_types[current_event].eventid < 0) {
    current_event++;
  }

  if (current_event >= TRACE_EVENT_MAX) {
    // No more valid event ids in the list
    *unavailable = 1;
  } else {
    // Report the event id and increase the counter
    *unavailable = 0;
    *event = event_types[current_event].eventid;
    current_event++;
  }
  return 0;
}





int posix_trace_eventtypelist_rewind (trace_id_t trid) {

  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
    return -EINVAL;
  }

  current_event = 0;
  return 0;
}







//
// 24.3.10. Manipulate trace event type sets
//
#ifdef _POSIX_TRACE_EVENT_FILTER
int posix_trace_eventset_empty (trace_event_set_t *set) {

  /* Not supported in Linux */
  return -EINVAL; 
}





int posix_trace_eventset_fill (trace_event_set_t *set,
			       int what) {

  /* Not supported in Linux */
  return -EINVAL; 
}






int posix_trace_eventset_add (trace_event_id_t event_id,
			      trace_event_set_t *set) {

  /* Not supported in Linux */
  return -EINVAL; 
}






int posix_trace_eventset_del (trace_event_id_t event_id,
			      trace_event_set_t *set) {

  /* Not supported in Linux */
  return -EINVAL; 
}






int posix_trace_eventset_complement(trace_event_set_t *set) {

  /* Not supported in Linux */
  return -EINVAL; 
}






int posix_trace_eventset_ismember (trace_event_id_t event_id,
				   const trace_event_set_t *set,
				   int *ismember) {

  /* Not supported in Linux */
  return -EINVAL; 
}





//
// 24.3.11. Manipulate trace event type sets
//

int posix_trace_set_filter (trace_id_t trid, 
			    const trace_event_set_t *set,
			    int how) {

  /* Not supported in Linux */
  return -EINVAL; 
}





int posix_trace_get_filter (trace_id_t trid, 
			    trace_event_set_t *set) {

  /* Not supported in Linux */
  return -EINVAL; 
}

#endif // #ifdef _POSIX_TRACE_EVENT_FILTER





//
// 24.3.12. Trace start and stop
//

int posix_trace_start (trace_id_t trid) { 

  /* Not supported in Linux */
  return -EINVAL; 
}





int posix_trace_stop (trace_id_t trid) {

  /* Not supported in Linux */
  return -EINVAL; 
}





//
// 24.3.14 Trace Log Management
//

int posix_trace_open (int file_desc, trace_id_t *trid) {

  int              magic;
  struct stat      statbuf;
  int              ret;


  // Error check
  if  (fstat(file_desc, &statbuf) == -1) {
    // "file_desc" is not valid
#ifdef _TRACE_DEBUG
    fprintf(stderr, "(open) file_desc is not open or not valid \n");
#endif
    return -EINVAL;
  }

  if (read(file_desc, &magic, sizeof(int)) == -1) {
#ifdef _TRACE_DEBUG
    fprintf(stderr, "(open) Cannot read initial magic in log \n");
#endif
    return -EINVAL;
  }

  if (magic != FLUSH_LOG_DONE) {
#ifdef _TRACE_DEBUG
    fprintf(stderr, "(open) Wrong initial magic in log \n");
#endif
    return -EINVAL;
  }


  *trid = pre_trid = file_desc;

  ret = read(pre_trid, &trid_attr, sizeof(trid_attr));
  if (ret < sizeof(trid_attr)) {
#ifdef _TRACE_DEBUG
    fprintf(stderr, "(open) Error while reading attr in log \n");
#endif
    return -EINVAL;
  }

  ret = read(pre_trid, &event_types, sizeof(event_types));
  if (ret < sizeof(event_types)) {
#ifdef _TRACE_DEBUG
    fprintf(stderr, "(open) Error while reading event types in log \n");
#endif
    return -EINVAL;
  }

  ret = read(pre_trid, &trid_status, sizeof(trid_status));
  if (ret < sizeof(trid_status)) {
#ifdef _TRACE_DEBUG
    fprintf(stderr, "(open) Error while reading status info in log \n");
#endif
    return -EINVAL;
  }

  first_event_pos = lseek(pre_trid, (off_t) 0, SEEK_CUR);
  current_event = 0;

  return 0;
}




int posix_trace_rewind (trace_id_t trid) {

  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
    return -EINVAL;
  }

  // Re-position the file pointer at the first event:
  lseek(pre_trid, (off_t) first_event_pos, SEEK_SET);
  return 0;
}





int posix_trace_close (trace_id_t trid) {

  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
    return -EINVAL;
  }
  
  pre_trid = -1;
  return 0;
}

















//
// 24.3.15. Retrieve the trace attributes or trace statuses
//
int posix_trace_get_attr (trace_id_t trid,
			  trace_attr_t *attr) {

  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
    return -EINVAL;
  }

  memcpy((void *) attr, (const void *) &trid_attr, sizeof(trace_attr_t));
  return 0;
}





int posix_trace_get_status (trace_id_t trid,
			    struct posix_trace_status_info *statusinfo) {

  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
    return -EINVAL;
  }

  memcpy((void *) statusinfo, (const void *) &trid_status, sizeof(posix_trace_status_info));
  return 0;

}





//
// 24.3.16. Retrieve a trace event
//
int posix_trace_getnext_event (trace_id_t trid,
			       struct posix_trace_event_info *event, 
			       void *data,
			       size_t num_bytes,
			       size_t *data_len,
			       int *unavailable) {
  int ret;
  int data_size;
  
  // Error checking:
  if (! TRACE_STREAM_IS_VALID(trid)) {
#ifdef _TRACE_DEBUG
    fprintf(stderr, "(getnext_event) Error: invalid trace_id %d (correct %d)\n",
	    (int) trid, (int) pre_trid);
#endif
    return -EINVAL;
  }

  // Cancellation test
  pthread_testcancel();

  ret = read(pre_trid, &event_read, sizeof(stream_event_t) - sizeof(void *));

  if (ret < (sizeof(stream_event_t) - sizeof(void *))) {
    *unavailable = 1;
  } else {
    *unavailable = 0;
    event->posix_event_id          = event_read.event_id;
    event->posix_pid               = 0;
    event->posix_prog_address      = event_read.prog_address;
    event->posix_truncation_status = (num_bytes < event_read.data_length? 
				      POSIX_TRACE_TRUNCATED_READ :
				      event_read.truncation_status);
    event->posix_thread_id         = event_read.thread_id;
    event->posix_timestamp         = timespec_from_ns(event_read.timestamp);

    if (event_read.data_length > 0) {
      data_size = (event_read.data_length < num_bytes ? 
                   event_read.data_length :
		   num_bytes);
      ret = read(pre_trid, data, data_size);
      *data_len = data_size;
      if (ret < data_size) {
	*unavailable = 1;
      }

      // If read less bytes than data_length, advance pointer until
      // next event
      if (num_bytes < event_read.data_length) {
	lseek(pre_trid, (off_t) (event_read.data_length - num_bytes),
	      SEEK_CUR);
      }
    }
  }

  return 0;
}

#endif






#ifdef _POSIX_TRACE_LINUXTRACE

static unsigned int    target_fifo;


//
// 24.3.13. Instrumenting application code
//
static inline int posix_trace_target_init(void) {

  static int started = 0;
  char       target_fifo_name[32];

  if (! started) {
    sprintf(target_fifo_name, "/dev/rtf%d", TRACE_TARGET_FIFO);
    if ((target_fifo = open(target_fifo_name, O_RDWR | O_SYNC)) == -1) {
      return -EINVAL;
    }
    started = 1;
  }
  
  return 0;
}





static inline void posix_trace_target_cleanup(void) {
  close(target_fifo);
}





int posix_trace_eventid_open (const char *event_name, 
			      trace_event_id_t *event_id) {

  fifo_request_t request;
  fifo_answer_t  answer;

  // Automatically open the target fifo (if necessary)
  if (posix_trace_target_init()) {
    return -EINVAL;
  }

  // Error check
  if (strlen(event_name) > TRACE_EVENT_NAME_MAX) {
    return -ENAMETOOLONG;
  }

  // Initialize the response message:
  answer.magic = NO_MAGIC;

  // Write the request:
  request.magic         = TRACE_FIFO_MAGIC;
  request.operation     = FIFO_TRACE_EVENTID_OPEN;
  strncpy(request.rdata.eventname, event_name, sizeof(request.rdata.eventname));
  write(target_fifo, (char *) &request, sizeof(fifo_request_t));

  // Read the answer:
  read(target_fifo, &answer, sizeof(fifo_answer_t));

  // Process the answer:
  if (answer.magic != TRACE_FIFO_MAGIC) {
    return -EINVAL;
  }

  *event_id = answer.data.eventid;
  return answer.return_value;
}





void posix_trace_event (trace_event_id_t event_id, 
			const void *data_ptr,      
			size_t data_len) {

  fifo_request_t request;
  int            calling_address;

  // Automatically open the target fifo (if necessary)
  if (posix_trace_target_init()) {
    return;
  }

  // Get the invoking address and store it in "calling_address"
  posix_trace_get_invoking_address(calling_address);

  // Write the request:
  request.magic         = TRACE_FIFO_MAGIC;
  request.operation     = FIFO_TRACE_EVENT;
  request.rdata.tracevent.eventid         = event_id;
  request.rdata.tracevent.data_len        = data_len;
  request.rdata.tracevent.calling_address = calling_address;
  memcpy(request.rdata.tracevent.data, data_ptr, data_len);
  write(target_fifo, (char *) &request, sizeof(fifo_request_t));
}


#endif






#endif /* #ifdef _POSIX_TRACE_LOGRET || _POSIX_TRACE_LINUXTRACE */
