/*
   FILE:							       	
	linux_analyzer.c
								       	
   DESCRIPTION:						
	      	
	Linux side the 4 API test of the POSIX trace support. 

	This Linux program retrieves all the information (and
	the user trace events) from the log stream created by the
	RT-Linux part (rtl_tasks.c).

					       				
   Copyright (C) 2004
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        December 2004

*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <trace.h>

#include "test_ptrace_api4.h"




int main(int argc, char *argv[]) {

  char         log_filename[256];
  char         str[TRACE_NAME_MAX];
  size_t       data_len;
  size_t       size;
  int          unavailable;
  int          pol;
  int          error;
  int          found_user_event;
  int          log_fd;
  int          number_of_events = 0;
  trace_event_id_t ev;
  trace_attr_t trace_attr;
  trace_id_t   trace_id;
  struct posix_trace_event_info event;
  struct posix_trace_status_info status;
  struct timespec  time;
  event_data_struct event_data;
  
  if (argc < 2) {
    fprintf(stderr, "%s: [BROK] Use: linux_analyzer logfile\n", TEST_NAME);
    exit(-1);
  } else if (argc == 2) {
    strncpy(log_filename, argv[1], sizeof(log_filename));
  }


  if ((log_fd = open(log_filename, O_RDONLY)) < 0) {
    fprintf(stderr, "%s: [BROK] linux_analyzer: Error while opening log file %s\n", 
	    TEST_NAME, log_filename);
    exit(-1);
  }

  error = posix_trace_open(log_fd, &trace_id);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_open returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }

  // Retrieve the attributes of this stream and check all the info:
  error = posix_trace_get_attr(trace_id, &trace_attr);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_get_attr returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }
  
  error = posix_trace_attr_getgenversion(&trace_attr, str);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getgenversion returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }

  error = posix_trace_attr_getname(&trace_attr, str);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getname returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }
  if (strcmp(str, MY_STREAM_NAME)) {
    fprintf(stderr, "%s: [BROK] Bad stream name found in log\n",
	    TEST_NAME);
    exit(-1);
  }

  error = posix_trace_attr_getcreatetime(&trace_attr, &time);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getcreatetime returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }


  error = posix_trace_attr_getclockres(&trace_attr, &time);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getclockres returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }
  
  error = posix_trace_attr_getstreamfullpolicy(&trace_attr, &pol);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getstreamfullpolicy returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }
  if (pol != MY_STREAM_POLICY) {
    fprintf(stderr, "%s: [BROK] Bad stream full policy found in log\n",
	    TEST_NAME);
    exit(-1);
  }

  error = posix_trace_attr_getmaxusereventsize(&trace_attr, 0, &size);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getmaxusereventsize returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }


  error = posix_trace_attr_getmaxsystemeventsize(&trace_attr, &size);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getmaxsystemeventsize returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }

  error = posix_trace_attr_getmaxdatasize(&trace_attr, &size);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getmaxdatasize returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }
  if (size != MY_EVENT_DATA_SIZE) {
    fprintf(stderr, "%s: [BROK] Bad event data size found in log\n",
	    TEST_NAME);
    exit(-1);
  }

  error = posix_trace_attr_getstreamsize(&trace_attr, &size);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_attr_getstreamsize returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }
  if (size != MY_STREAM_SIZE) {
    fprintf(stderr, "%s: [BROK] Bad stream size found in log\n",
	    TEST_NAME);
    exit(-1);
  }




  // -------------------------------------------------------------------------
  // Print the full list of event types available in the stream:

  error = posix_trace_eventtypelist_rewind(trace_id);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_eventtypelist_rewind returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }

  error = posix_trace_eventtypelist_getnext_id (trace_id, &ev, &unavailable);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_eventtypelist_getnext_id returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }


  /* Taverse the full list of event types looking for the user event type
     defined before
  */
  found_user_event = 0;
  while( ! unavailable && ! error) {
    
    error = posix_trace_eventid_get_name (trace_id, ev, str);
    if (error != 0) {
      fprintf(stderr, "%s: [BROK] posix_trace_get_name returned %d\n",
	      TEST_NAME, error);
      exit(-1);
    }

    if (! strcmp(MY_USER_EVENT_NAME, str)) {
      found_user_event = 1;
    }

    // Get the next one:
    error = posix_trace_eventtypelist_getnext_id (trace_id, &ev, 
						  &unavailable);
  }

  if (! found_user_event) {
    fprintf(stderr, "%s: [BROK] Event type list doesn't contain the user event type\n",
	    TEST_NAME);
    exit(-1);
  }


  // Get the stream status
  error = posix_trace_get_status(trace_id, &status);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_get_status returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }

  /*
    Retrieve all the events in the stream, checking if there are the expected
     number of them and their contents are exactly what traced
  */

  number_of_events = 0;

  error = posix_trace_getnext_event(trace_id, &event, 
				    (void *) &event_data, 
				    sizeof(event_data),
				    &data_len, &unavailable);
  
  while (! error && ! unavailable) {

    posix_trace_eventid_get_name (trace_id, event.posix_event_id, str);

    if (strcmp(MY_USER_EVENT_NAME, str)) {
      fprintf(stderr, "%s: [BROK] Stream contains an event of wrong type\n",
	      TEST_NAME);
      exit(-1);
    }

    if (strcmp(MY_STRING, (char *) event_data.str)) {
      fprintf(stderr, "%s: [BROK] Event in stream has wrong string data\n",
	      TEST_NAME);
      exit(-1);
    }
    
    if (event_data.serial != number_of_events) {
      fprintf(stderr, "%s: [BROK] Event in stream has wrong serial data\n",
	      TEST_NAME);
      exit(-1);
    }

    number_of_events ++;
    error = posix_trace_getnext_event(trace_id, &event, 
				      (void *) &event_data, 
				      sizeof(event_data),
				      &data_len, &unavailable);
  }

  if (number_of_events != NUMBER_OF_TRACED_EVENTS) {
      fprintf(stderr, "%s: [BROK] Wrong number of events in the stream (expected %d, got %d)\n",
	      TEST_NAME, NUMBER_OF_TRACED_EVENTS, number_of_events);
      exit(-1);
  }

  /* 
     Now rewind the stream and check if the first retrieved event
     is indeed the first traced one (serial == 0)
  */
  error = posix_trace_rewind(trace_id);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_rewind returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }

  error = posix_trace_getnext_event(trace_id, &event, 
				    (void *) &event_data, 
				    sizeof(event_data),
				    &data_len, &unavailable);
  
  if (error || unavailable) {
      fprintf(stderr, "%s: [BROK] After rewind, no events in stream\n",
	      TEST_NAME);
      exit(-1);
  }

  if (event_data.serial != 0) {
    fprintf(stderr, "%s: [BROK] After rewind, first event is wrong\n",
	    TEST_NAME);
    exit(-1);
  }

  /* Now close the stream and we're done! */

  error = posix_trace_close(trace_id);
  if (error != 0) {
    fprintf(stderr, "%s: [BROK] posix_trace_close returned %d\n",
	    TEST_NAME, error);
    exit(-1);
  }

  // If reach here, test is successful. Return 0.
  exit(0);

}
