/*
   FILE:
						       	
	test_ptrace_stress1_module.c
								       	
   DESCRIPTION:							       	

        Stress test 1 for POSIX Tracing in RT-Linux.

	This test checks the limits of size of streams and events
	that can be created in RT-Linux, by creating streams with
	increasingly size.
 	
	For each stream size (from 512Kb to 64Mb, doubling each time),
	the corresponding event size is set to be 16 times smaller
	(that is, up to 16 events would fit in the stream). For very
	big stream sizes, this also means really big events.

					       				
   Copyright (C) 2004
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        September 2004

*/

#include <rtl.h>
#include <time.h>
#include <trace.h>
#include <asm/errno.h>


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andres Terrasa");
MODULE_DESCRIPTION("Stress Test 1 for POSIX Tracing");


#define TEST_NAME "ptrace_test_stress1"


#define MIN_STREAM_SIZE         512 
#define MAX_STREAM_SIZE    67108864 /* 64 Mb */



int init_module(void) {

  trace_attr_t   attr;
  int            error;
  trace_id_t       trid;
  size_t         stream_size;
  size_t         event_data_size;

  
  // Initialize a trace attribute objects:
  posix_trace_attr_init(&attr);


  /*
    Now enter a loop where streams are created with increasing
    size and bigger events, until -ENOMEN is returned or size
    reaches 64Mb 
  */

  stream_size     = MIN_STREAM_SIZE;
  event_data_size = (int) (stream_size / 16);
  
  posix_trace_attr_setstreamsize(&attr, stream_size);
  posix_trace_attr_setmaxdatasize(&attr, event_data_size);
  error = posix_trace_create(0, &attr, &trid);

  while (! error && stream_size < MAX_STREAM_SIZE) {
    posix_trace_shutdown(trid);
    stream_size *= 2;
    event_data_size = (int) (stream_size / 16);

    posix_trace_attr_setstreamsize(&attr, stream_size);
    posix_trace_attr_setmaxdatasize(&attr, event_data_size);
    error = posix_trace_create(0, &attr, &trid);
  }

  if (stream_size < MAX_STREAM_SIZE) {
    if (error == -ENOMEM) {
      rtl_printf("%s: [BROK]:Cannot create stream: too big (stream %d, event data %d)\n", 
		 TEST_NAME, stream_size, event_data_size);
    } else {
      rtl_printf("%s: [BROK]:Cannot create stream: too small or other bad arguments\n", 
		 TEST_NAME, stream_size, event_data_size);
    }
    return -1;
  }

  posix_trace_shutdown(trid);

  /* No errors. All tests were fine. Return 0. */
  rtl_printf("%s: PASS :Memory stress test (512Kb to 64Mb streams) passed\n",
	     TEST_NAME);  
  return 0;

}



void cleanup_module(void) {

}
